@@ -43,6 +43,14 @@ void changeToIndirectSend(G4_INST *inst, G4_Declare *s0Var, int totalRegs,
                           IR_Builder &builder, bool isLargeGRF) {
   // Change the send instruction to sendi
   G4_InstSend *Send = inst->asSendInst();
+  if (Send->opcode() == G4_sendg || Send->opcode() == G4_sendgc) {
+    G4_SendgDesc *desc = (G4_SendgDesc *)inst->getMsgDesc();
+    desc->setSrc0Len(totalRegs);
+    desc->setSrc1Len(0);
+    if (isLargeGRF) {
+      Send->setSendgx();
+    }
+  } else {
     G4_SendDescRaw *desc = Send->getMsgDescRaw();
     desc->setExtMessageLength(0);
     G4_Operand *msgDesc = inst->getSrc(2);
@@ -52,6 +60,7 @@ void changeToIndirectSend(G4_INST *inst, G4_Declare *s0Var, int totalRegs,
     descImm |= totalRegs << 25;
     G4_Imm *msgDescImm = builder.createImm(descImm, Type_UD);
     inst->setSrc(msgDescImm, 2);
+  }
 
   // Replace source 0 with scalar register
   G4_SrcRegRegion *headerOpnd =
@@ -64,370 +73,195 @@ void changeToIndirectSend(G4_INST *inst, G4_Declare *s0Var, int totalRegs,
   inst->setSrc(payloadToUse, 1);
 }
 
-// Check if current instruction is the candidate of sendi.
-// Recorded as candidate.
-bool SRSubPass::isSRCandidate(G4_INST *inst, regCandidates &dstSrcRegs) {
-  if (!inst->isSend()) {
+bool SRSubPassAfterRA::isLargeGRFOpnd(G4_Operand *opnd) {
+  if (!opnd || opnd->isNullReg() || !opnd->isGreg()) {
+    return false;
+  }
+  return opnd->getLinearizedEnd() > (256 * builder.numEltPerGRF<Type_UB>() - 1);
+}
+
+bool SRSubPassAfterRA::canRegisterBeEncoded(G4_Operand *opnd) {
+  // In some platforms, mov is not removable and dst register > r255, cannot be
+  // gather instruction
+  return !isLargeGRFOpnd(opnd);
+}
+
+bool SRSubPassAfterRA::isRemoveAble(G4_INST *i) {
+  if (i->opcode() != G4_mov) {
     return false;
   }
 
-  // The payload size is less than 2 GRFs, no need to use sendi
-  if (inst->getMsgDesc()->getSrc1LenRegs() < 2) {
+  if (i->getPredicate() || i->getCondMod() || i->getSaturate()) {
     return false;
   }
 
-  int movInstNum = 0;
-  int32_t firstDefID = 0x7FFFFFFF; // the ID of the first instruction define the
-                                   // payload of the send
-  std::list<USE_DEF_NODE> SRCandidate;
-  SRCandidate.clear();
-  bool src0Out = false;
-  for (auto I = inst->def_begin(), E = inst->def_end(); I != E; ++I) {
-    auto &&def = *I;
+  // The instruction is only used for payload preparation.
+  if (i->use_size() != 1) {
+    return false;
+  }
 
-    if (def.second != Opnd_src1 && def.second != Opnd_src0) {
-      continue;
-    }
+  G4_DstRegRegion *dstRgn = i->getDst();
+  G4_Operand *src = i->getSrc(0);
 
-    if (def.first->opcode() == G4_mov) {
-      movInstNum++;
-    }
+  if (!src->isSrcRegRegion()) {
+    return false;
+  }
 
-    // The instruction is only used for payload preparation.
-    if (def.first->use_size() != 1) {
+  // Not GRF source
+  if (!(src->getBase()->isRegVar() &&
+        (src->getBase()->asRegVar()->getDeclare()->getRegFile() == G4_GRF ||
+         src->getBase()->asRegVar()->getDeclare()->getRegFile() == G4_INPUT))) {
+    return false;
+  }
+
+  // The src region is scalar
+  if (src->isScalarSrc()) {
+    return false;
+  }
+
+  G4_SrcRegRegion *srcRgn = src->asSrcRegRegion();
+
+  if (srcRgn->hasModifier()) {
+    return false;
+  }
+
+  // dst GRF aligned and contigous
+  if (dstRgn->getSubRegOff() || dstRgn->getHorzStride() != 1) {
+    return false;
+  }
+
+  // src GRF aligned and contigous
+  if (srcRgn->getSubRegOff() ||
+      !srcRgn->getRegion()->isContiguous(i->getExecSize())) {
+    return false;
+  }
+
+  // The src region is not packed
+  if (!srcRgn->getRegion()->isPackedRegion()) {
+    return false;
+  }
+
+  // No type conversion
+  if (dstRgn->getType() != src->getType()) {
+    return false;
+  }
+
+  // If the destination operand size is less than 1 GRF
+  if ((dstRgn->getLinearizedEnd() - dstRgn->getLinearizedStart() + 1) <
+      builder.getGRFSize()) {
+    return false;
+  }
+
+  // If the source operand size is less than 1 GRF
+  if ((src->getLinearizedEnd() - src->getLinearizedStart() + 1) <
+      builder.getGRFSize()) {
+    return false;
+  }
+
+  // GRF Alignment with physical register assigned
+  if (dstRgn->getLinearizedStart() % builder.getGRFSize() != 0 ||
+      src->getLinearizedStart() % builder.getGRFSize() != 0) {
+    return false;
+  }
+
+  // Move to self instruction will be removed by following pass. So cannot
+  // be counted.
+  if (dstRgn->getLinearizedStart() == src->getLinearizedStart()) {
+    return false;
+  }
+
+  // It's not global define
+  if (!(builder.getIsKernel() && kernel.fg.getNumBB() == 1)) {
+    if (kernel.fg.globalOpndHT.isOpndGlobal(dstRgn) &&
+        !dstRgn->getTopDcl()->getIsBBLocal()) {
       return false;
     }
-
-    // It's not global define
-    if (kernel.fg.globalOpndHT.isOpndGlobal(def.first->getDst())) {
-      if (def.second == Opnd_src1) {
-        return false;
-      } else { //src0
-        src0Out = true;
-        continue;
-      }
-    }
-
-    SRCandidate.push_back(def);
-    firstDefID = std::min(firstDefID, def.first->getLocalId());
   }
 
-  // Check if there are enough mov instructions to be removed.
-  if (movInstNum < 2) {
+  // In some platforms, mov is not removeable if the src register > r255
+  if (!canRegisterBeEncoded(src)) {
     return false;
   }
 
-  unsigned short GRFSize = builder.getGRFSize();
-  unsigned short GRFNum = builder.kernel.getNumRegTotal();
-  BitSet definedReg(GRFNum, false);
-  for (auto I = SRCandidate.begin(), E = SRCandidate.end(); I != E; ++I) {
-    G4_INST *cInst = (*I).first;
-    Gen4_Operand_Number opndNum = (*I).second;
-    G4_Operand *dst = cInst->getDst();
-    // GRF size aligned
-    if ((dst->getLinearizedEnd() - dst->getLinearizedStart() + 1) % GRFSize !=
-        0) {
-      return false;
-    }
-
-    for (unsigned i = dst->getLinearizedStart() / GRFSize; i < dst->getLinearizedEnd() / GRFSize; i ++ ) {
-      // r4, r5, r6, r7 are multiple defined
-      // (W) send.ugm(1 | M0) r4 r36 null:0 0x85000000:a0.2 0x4240F500 {$18} //
-      // wr:1+0, rd:4; load.ugm.d32x64t.a32.flat[A-0x7B000] fill from
-      // offset[40*64] of COAL_FILL_3164; $954
-      // mov(32 | M0) r4 .0 < 1 > : ud r10 .0 < 1;1,0> : ud{Compacted, $18.dst}
-      // mov(32 | M0) r6 .0 < 1 > : ud r8 .0 < 1;1,0> :  ud{Compacted}
-      // (W) send.ugm(1 | M0) null r36 r4 : 4 0x85000000:a0.2 0x4200F504 {I @1,
-      // $19} // wr:1+4, rd:0; store.ugm.d32x64t.a32.flat[A-0x7B000]   spill to
-      // offset[40*64] of COAL_FILL_3164; $954
-      if (definedReg.isSet(i) && opndNum == Opnd_src1) { //No multiple define in src1
-        return false;
-      } else {
-        definedReg.set(i, true);
-      }
-    }
-    // The payload preparation instruction may be mov,or other instsruction
-    // For mov instruction, use the source register and remove the mov
-    // instruction For non-mov instruction, use the dst register directly
-    // without remove the instruction
-    //  or (8)              r15.0 < 1 > :d  r114.0 < 1; 1, 0 > : d  r2.0 < 1; 1, 0 > : d
-    //  mov(8)              r13.0 < 1 > :f  r21.0 < 1; 1, 0 > : f
-    //  mov(8)              r14.0 < 1 > :f  r18.0 < 1; 1, 0 > : f
-    //  mov(8)              r16.0 < 1 > :f  r55.0 < 1; 1, 0 > : f
-    //  mov(8)              r12.0 < 1 > :d  r111.0 < 1; 1, 0 > : d
-    //  sends(8)            null : ud  r112  r12 0x14c : ud 0x2035091 :
-    //  ud //typed surface write, resLen=0, msgLen=1, extMsgLen=5
-    if (cInst->opcode() == G4_mov) {
-      G4_Operand *src = cInst->getSrc(0);
-      unsigned short dstRegLB = dst->getLinearizedStart() / GRFSize;
-      unsigned short dstRegRB = dst->getLinearizedEnd() / GRFSize;
-      unsigned short srcRegLB = src->getLinearizedStart() / GRFSize;
-      unsigned short srcRegRB = src->getLinearizedEnd() / GRFSize;
-
-      // GRF aligned
-      if (dst->getLinearizedStart() % GRFSize != 0 ||
-          src->getLinearizedStart() % GRFSize != 0) {
-        return false;
-      }
-
-      // In case src region is not consistent with dst region
-      if (srcRegRB - srcRegLB != dstRegRB - dstRegLB) {
-        return false;
-      }
-    } else {
-      // GRF aligned
-      if (dst->getLinearizedStart() % GRFSize != 0) {
-        return false;
-      }
-    }
-  }
-
-  for (auto I = SRCandidate.begin(), E = SRCandidate.end(); I != E; ++I) {
-    G4_INST *cInst = (*I).first;
-    Gen4_Operand_Number opndNum = (*I).second;
-
-    if (src0Out && opndNum == Opnd_src0) {
-      continue;
-    }
-
-    G4_Operand *dst = cInst->getDst();
-    unsigned short dstRegLB = dst->getLinearizedStart() / GRFSize;
-    unsigned short dstRegRB = dst->getLinearizedEnd() / GRFSize;
-
-    if (cInst->opcode() == G4_mov) {
-      G4_Operand *src = cInst->getSrc(0);
-      unsigned short srcRegLB = src->getLinearizedStart() / GRFSize;
-      for (unsigned short reg = dstRegLB; reg <= dstRegRB; reg++) {
-        regMap regPair(cInst->getLocalId(), reg,
-                       srcRegLB + reg - dstRegLB);
-        dstSrcRegs.dstSrcMap.push_back(regPair);
-        if (opndNum == Opnd_src0)
-        {
-            dstSrcRegs.includeSrc0 = true;
-        }
-      }
-    } else // For none-mov instruction, we will use the dst directly,
-           // instruction will not be erased.
-    {
-      for (unsigned short reg = dstRegLB; reg <= dstRegRB; reg++) {
-        regMap regPair(cInst->getLocalId(), reg, reg);
-        dstSrcRegs.dstSrcMap.push_back(regPair);
-      }
-    }
-  }
-
-  // Record the smallest local ID of the define instructions, for the valid
-  // candidate scan.
-  dstSrcRegs.firstDefID = firstDefID;
-
-  // Sort according to the register order in the original payload
-  std::sort(dstSrcRegs.dstSrcMap.begin(), dstSrcRegs.dstSrcMap.end(),
-            regSortCompare);
-
   return true;
 }
 
-// Replace the send instruction with the payload of
-// Insert the scalar register intialization mov instructions.
-bool SRSubPass::replaceWithSendi(G4_BB *bb, INST_LIST_ITER instIter,
-                                 std::vector<regMap> &dstSrcRegs, bool src0Mov) {
-  G4_INST *inst = *instIter;
-  int numRows = dstSrcRegs.size();
-  int totalRegs = numRows;
-  vASSERT(totalRegs <= MAXIMAL_S0_SRC0_GRF_LENGTH);
-  uint64_t value = 0;
-  uint64_t value1 = 0;
-  int value_size = 0;
-  int value1_size = 0;
+// mov (16)             r81.0<1>:f  0x8:f // $52:&54:
+// mov (16|M16)         r89.0<1>:f  0x8:f // $53:&55:
+// mov (16)             r82.0<1>:f  0x0:f // $54:&56:
+// mov (16|M16)         r90.0<1>:f  0x0:f // $55:&57:
+// mov (16)             r83.0<1>:f  0x0:f // $56:&58:
+// mov (16|M16)         r91.0<1>:f  0x0:f // $57:&59:
+// mov (16)             r84.0<1>:f  0x0:f // $58:&60:
+// mov (16|M16)         r92.0<1>:f  0x0:f // $59:&61:
+// mov (16)             r85.0<1>:f  0x0:f // $60:&62:
+// mov (16|M16)         r93.0<1>:f  0x0:f // $61:&63:
+// mov (16)             r86.0<1>:f  0x0:f // $62:&64:
+// mov (16|M16)         r94.0<1>:f  0x0:f // $63:&65:
+// mov (16)             r87.0<1>:f  0x0:f // $64:&66:
+// mov (16|M16)         r95.0<1>:f  0x0:f // $65:&67:
+// mov (16)             r88.0<1>:f  0x0:f // $66:&68:
+// mov (16|M16)         r96.0<1>:f  0x0:f // $67:&69:
+// ==>
+// mov (16)             r81.0<1>:f  0x8:f // $52:&54:
+// mov (16|M16)         r89.0<1>:f  0x8:f // $53:&55:
+// mov (16)             r82.0<1>:f  0x0:f // $54:&56:
+// mov (16|M16)         r90.0<1>:f  0x0:f // $55:&57:
+//
+// Reuse r81, r89, r82, r90 in the gather send
+G4_INST *SRSubPassAfterRA::getRemoveableImm(G4_INST *inst,
+                                            std::vector<G4_INST *> &immMovs) {
+  // The instruction is only used for payload preparation.
+  if (inst->use_size() != 1) {
+    return (G4_INST *)nullptr;
+  }
 
-  if (!src0Mov) {
-    G4_Operand *src0 = inst->getSrc(0);
-    unsigned short GRFSize = builder.getGRFSize();
-    unsigned short src0RegLB = src0->getLinearizedStart() / GRFSize;
-    unsigned short src0RegRB = src0->getLinearizedEnd() / GRFSize;
-    totalRegs += src0RegRB - src0RegLB + 1;
+  if (inst->opcode() != G4_mov) {
+    return (G4_INST *)nullptr;
+  }
 
-    // Original src0
-    for (int i = 0; i < (src0RegRB - src0RegLB + 1); i++, value_size++) {
-      value |= ((uint64_t)src0RegLB + i) << (8 * i);
+  G4_DstRegRegion *dst = inst->getDst();
+  // dst GRF aligned and contigous
+  if (dst->getSubRegOff() || dst->getHorzStride() != 1) {
+    return (G4_INST *)nullptr;
+  }
+
+  if (kernel.fg.globalOpndHT.isOpndGlobal(dst)) {
+    return (G4_INST *)nullptr;
+  }
+
+  // GRF Alignment with physical register assigned
+  if (dst->getLinearizedStart() % builder.getGRFSize() != 0) {
+    return (G4_INST *)nullptr;
+  }
+
+  // If the destination operand size is less than 1 GRF
+  if ((dst->getLinearizedEnd() - dst->getLinearizedStart() + 1) <
+      builder.getGRFSize()) {
+    return (G4_INST *)nullptr;
+  }
+
+  G4_Operand *src = inst->getSrc(0);
+  int64_t imm = src->asImm()->getImm();
+  for (size_t i = 0; i < immMovs.size(); i++) {
+    G4_INST *imov = immMovs[i];
+    G4_Operand *isrc = imov->getSrc(0);
+    int64_t iimm = isrc->asImm()->getImm();
+    if (imm == iimm &&
+        src->getType() == isrc->getType() && // Same value and same type
+        inst->getDst()->getType() ==
+            imov->getDst()->getType() && // Same dst type
+        inst->getDst()->asDstRegRegion()->getHorzStride() ==
+            imov->getDst()->asDstRegRegion()->getHorzStride() && // Same region
+        inst->getExecSize() == imov->getExecSize() &&     // Same execution size
+        inst->getMaskOffset() == imov->getMaskOffset()) { // Same mask offset
+      return imov;
     }
   }
 
-  // Plusing the src1, totally 8 element at most
-  int i = 0;
-  for (; i < numRows && value_size < 8; i++, value_size++) {
-    value |= ((uint64_t)dstSrcRegs[i].srcReg) << 8 * value_size;
-  }
-  vASSERT(value_size <= 8);
+  immMovs.push_back(inst);
 
-  // The remaining data. May more than 8 GRFs, so will need two mov instructions
-  for (int j = i; j < numRows; j++, value1_size++) {
-    value1 |= ((uint64_t)dstSrcRegs[j].srcReg) << 8 * (j - i);
-  }
-
-  // Initialize the scalar registers.
-  uint16_t UQNum = totalRegs > (TypeSize(Type_UQ) / TypeSize(Type_UB)) ? 2 : 1;
-  G4_Declare *s0Var = builder.createScalar(UQNum, "S0_");
-  s0Var->getRegVar()->setPhyReg(builder.phyregpool.getScalarReg(), 0);
-  G4_Imm *src = builder.createImm(value, Type_UQ);
-  G4_DstRegRegion *dst =
-      builder.createDst(s0Var->getRegVar(), 0, 0, 1, Type_UQ);
-  G4_INST *movInst =
-      builder.createMov(g4::SIMD1, dst, src, InstOpt_WriteEnable, false);
-  bb->insertBefore(instIter, movInst);
-
-  // Second initialization if more than 8 GRFs
-  if (value1_size) {
-    G4_Imm *src = builder.createImm(value1, Type_UQ);
-    G4_DstRegRegion *dst =
-        builder.createDst(s0Var->getRegVar(), 0, 1, 1, Type_UQ);
-    G4_INST *movInst =
-        builder.createMov(g4::SIMD1, dst, src, InstOpt_WriteEnable, false);
-    bb->insertBefore(instIter, movInst);
-  }
-
-  changeToIndirectSend(inst, s0Var, totalRegs, builder, false);
-
-  return true;
-}
-
-// Subsititue with scalar regsiter
-void SRSubPass::SRSub(G4_BB *bb) {
-  bb->resetLocalIds();
-
-  class CmpFirstDef {
-  public:
-    bool operator()(G4_INST *first, G4_INST *second) const {
-      return first->getLocalId() < second->getLocalId();
-    }
-  };
-  std::map<G4_INST *, regCandidates, CmpFirstDef> candidates;
-  std::map<G4_INST *, regCandidates, CmpFirstDef>::iterator candidatesIt;
-
-  INST_LIST_ITER ii = bb->begin(), iend(bb->end());
-  while (ii != iend) {
-    G4_INST *inst = *ii;
-
-    regCandidates dstSrcRegs;
-    if (!isSRCandidate(inst, dstSrcRegs)) {
-      ii++;
-      continue;
-    }
-    candidates.insert(std::make_pair(inst, dstSrcRegs));
-    ii++;
-  }
-
-  // If the registser is redefined before the send, the substitution needs be
-  // cancelled. Such as r9 in following instructions, which is redefined in sel
-  // instruction.
-  // mov(8)             r6.0< 1 > :f  r9.0 < 1; 1, 0 > : f
-  // sel(8) (ge)f0.0    r9.0< 1 > :f  r57.0 < 1; 1, 0 > : f
-  // mov(8)             r3.0< 1 > :f  r79.0 < 1; 1, 0 > : f
-  // mov(8)             r4.0 < 1 > : f  r27.0 < 1; 1, 0 > : f
-  // mov(8)             r5.0 < 1 > : f  r77.0 < 1; 1, 0 > : f
-  // mov(8)             r2.0 < 1 > : d  r111.0 < 1; 1, 0 > : d
-  // sends(8)           null : ud r112 r2 0x14c : ud 0x2035092 : ud //typed
-  // surface write, resLen=0, msgLen=1, extMsgLen=5
-  INST_LIST_RITER ri = bb->rbegin(), rend(bb->rend());
-  while (ri != rend) {
-    G4_INST *inst = *ri;
-    candidatesIt = candidates.find(inst);
-    if (candidatesIt != candidates.end()) {
-      bool overwrite = false;
-      INST_LIST_RITER scan_ri = ri;
-      scan_ri++;
-      G4_INST *rInst = *scan_ri;
-      while (rInst->getLocalId() > candidates[inst].firstDefID) {
-        G4_Operand *dst = rInst->getDst();
-        if (dst && !dst->isNullReg()) {
-          G4_VarBase *base = dst->getBase();
-          G4_VarBase *phyReg =
-              (base->isRegVar()) ? base->asRegVar()->getPhyReg() : base;
-
-          if (phyReg->getKind() == G4_VarBase::VK_phyGReg) {
-            // If the register reuse happened in the mov instruction which will
-            // be removed. Or the instruction whose dst will be accessed through
-            // s0. It's not real reuse. Such as the r8 in following
-            // instructions.
-            //clang-format off
-            // mov(8)  r9.0 <1>:f  r8.0<1; 1, 0>:f
-            // mov(8)  r10.0<1>:f  r3.0<8; 8, 1>:f
-            // mov(8)  r11.0<1>:f  r3.0<8; 8, 1>:f
-            // mov(8)  r7.0 <1>:d  r5.0<1; 1, 0>:d
-            // mov(8)  r8.0 <1>:f  r3.0<1; 1, 0>:f
-            // sends(8) null:ud r112 r7 0x14c : ud 0x2035093 : ud // typed surface write, resLen=0, msgLen=1, extMsgLen=5
-            //clang-format on
-            bool sameInst = false;
-            for (int i = 0; i < (int)candidates[inst].dstSrcMap.size(); i++) {
-              if (rInst->getLocalId() ==
-                  candidates[inst].dstSrcMap[i].localID) {
-                sameInst = true;
-                break;
-              }
-            }
-
-            if (!sameInst) {
-              unsigned short GRFSize = builder.getGRFSize();
-              G4_Operand *dst = rInst->getDst();
-              unsigned short dstRegLB = dst->getLinearizedStart() / GRFSize;
-              unsigned short dstRegRB = dst->getLinearizedEnd() / GRFSize;
-              for (int i = 0; i < (int)candidates[inst].dstSrcMap.size(); i++) {
-                int srcReg = candidates[inst].dstSrcMap[i].srcReg;
-                if (!(srcReg < dstRegLB || srcReg > dstRegRB)) {
-                  // Register is reused.
-                  overwrite = true;
-                  break;
-                }
-              }
-            }
-          }
-        }
-        scan_ri++;
-        if (scan_ri == rend) {
-          break;
-        }
-        rInst = *scan_ri;
-      }
-      if (overwrite) {
-        candidates.erase(candidatesIt);
-      }
-    }
-
-    ri++;
-  }
-
-  for (candidatesIt = candidates.begin(); candidatesIt != candidates.end();
-       candidatesIt++) {
-    G4_INST *inst = candidatesIt->first;
-
-    for (auto I = inst->def_begin(), E = inst->def_end(); I != E; ++I) {
-      auto &&def = *I;
-      if (def.first->opcode() == G4_mov) // Only mov instruction will be erased
-      {
-        def.first->markDead();
-      }
-    }
-  }
-
-  // Replace the send instruction with sendi
-  // Remove the mov instructions that marked as dead
-  INST_LIST_ITER iter;
-  for (iter = bb->begin(); iter != bb->end();) {
-    G4_INST *inst = *iter;
-    INST_LIST_ITER curIter = iter;
-    iter++;
-
-    candidatesIt = candidates.find(inst);
-    if (candidatesIt != candidates.end()) {
-      replaceWithSendi(bb, curIter, candidates[inst].dstSrcMap,
-                       candidates[inst].includeSrc0);
-    }
-    if (inst->isDead()) {
-      bb->erase(curIter);
-    }
-    // FIXME: after replacing with sendi and scalar register mov, the local data
-    // flow analysis need be able to capture the change.
-  }
+  return (G4_INST *)nullptr;
 }
 
 // Check if current instruction is the candidate of sendi.
@@ -438,7 +272,7 @@ bool SRSubPassAfterRA::isSRCandidateAfterRA(G4_INST *inst,
     return false;
   }
 
-  if (!inst->asSendInst()->getMsgDescRaw()) {
+  if (!inst->asSendInst()->getMsgDescRaw() && !inst->isSendg()) {
     return false;
   }
 
@@ -480,6 +314,12 @@ bool SRSubPassAfterRA::isSRCandidateAfterRA(G4_INST *inst,
     return false;
   }
 
+  if (inst->getDst() && !inst->getDst()->isNullReg()) {
+    // In some platforms, gather send doesn't support register > r255
+    if (!canRegisterBeEncoded(inst->getDst())) {
+      return false;
+    }
+  }
 
   int movInstNum = 0;
   int32_t firstDefID = 0x7FFFFFFF; // the ID of the first instruction define the
@@ -494,180 +334,6 @@ bool SRSubPassAfterRA::isSRCandidateAfterRA(G4_INST *inst,
     }
 
     G4_INST *movInst = def.first;
-    auto isRemoveAble = [this](G4_INST *i) {
-      if (i->opcode() != G4_mov) {
-        return false;
-      }
-
-      if (i->getPredicate() || i->getCondMod() || i->getSaturate()) {
-        return false;
-      }
-
-      // The instruction is only used for payload preparation.
-      if (i->use_size() != 1) {
-        return false;
-      }
-
-      G4_DstRegRegion *dstRgn = i->getDst();
-      G4_Operand *src = i->getSrc(0);
-
-      if (!src->isSrcRegRegion()) {
-        return false;
-      }
-
-      // Not GRF source
-      if (!(src->getBase()->isRegVar() &&
-            (src->getBase()->asRegVar()->getDeclare()->getRegFile() == G4_GRF ||
-             src->getBase()->asRegVar()->getDeclare()->getRegFile() == G4_INPUT))) {
-        return false;
-      }
-
-      // The src region is scalar
-      if (src->isScalarSrc()) {
-        return false;
-      }
-
-      G4_SrcRegRegion *srcRgn = src->asSrcRegRegion();
-
-      if (srcRgn->hasModifier()) {
-        return false;
-      }
-
-      //dst GRF aligned and contigous
-      if (dstRgn->getSubRegOff() ||
-          dstRgn->getHorzStride() != 1) {
-        return false;
-      }
-
-      // src GRF aligned and contigous
-      if (srcRgn->getSubRegOff() ||
-          !srcRgn->getRegion()->isContiguous(i->getExecSize())) {
-        return false;
-      }
-
-      // The src region is not packed
-      if (!srcRgn->getRegion()->isPackedRegion()) {
-        return false;
-      }
-
-      //No type conversion
-      if (dstRgn->getType() != src->getType()) {
-        return false;
-      }
-
-      // If the destination operand size is less than 1 GRF
-      if ((dstRgn->getLinearizedEnd() - dstRgn->getLinearizedStart() + 1) <
-          builder.getGRFSize()) {
-        return false;
-      }
-
-      // If the source operand size is less than 1 GRF
-      if ((src->getLinearizedEnd() - src->getLinearizedStart() + 1) <
-          builder.getGRFSize()) {
-        return false;
-      }
-
-      //GRF Alignment with physical register assigned
-      if (dstRgn->getLinearizedStart() % builder.getGRFSize() != 0 ||
-          src->getLinearizedStart() % builder.getGRFSize() != 0) {
-        return false;
-      }
-
-      // Move to self instruction will be removed by following pass. So cannot
-      // be counted.
-      if (dstRgn->getLinearizedStart() == src->getLinearizedStart()) {
-        return false;
-      }
-
-      // It's not global define
-      if (!(builder.getIsKernel() && kernel.fg.getNumBB() == 1)) {
-        if (kernel.fg.globalOpndHT.isOpndGlobal(dstRgn) && !dstRgn->getTopDcl()->getIsBBLocal()) {
-          return false;
-        }
-      }
-
-      return true;
-    };
-
-    // mov (16)             r81.0<1>:f  0x8:f // $52:&54:
-    // mov (16|M16)         r89.0<1>:f  0x8:f // $53:&55:
-    // mov (16)             r82.0<1>:f  0x0:f // $54:&56:
-    // mov (16|M16)         r90.0<1>:f  0x0:f // $55:&57:
-    // mov (16)             r83.0<1>:f  0x0:f // $56:&58:
-    // mov (16|M16)         r91.0<1>:f  0x0:f // $57:&59:
-    // mov (16)             r84.0<1>:f  0x0:f // $58:&60:
-    // mov (16|M16)         r92.0<1>:f  0x0:f // $59:&61:
-    // mov (16)             r85.0<1>:f  0x0:f // $60:&62:
-    // mov (16|M16)         r93.0<1>:f  0x0:f // $61:&63:
-    // mov (16)             r86.0<1>:f  0x0:f // $62:&64:
-    // mov (16|M16)         r94.0<1>:f  0x0:f // $63:&65:
-    // mov (16)             r87.0<1>:f  0x0:f // $64:&66:
-    // mov (16|M16)         r95.0<1>:f  0x0:f // $65:&67:
-    // mov (16)             r88.0<1>:f  0x0:f // $66:&68:
-    // mov (16|M16)         r96.0<1>:f  0x0:f // $67:&69:
-    // ==>
-    // mov (16)             r81.0<1>:f  0x8:f // $52:&54:
-    // mov (16|M16)         r89.0<1>:f  0x8:f // $53:&55:
-    // mov (16)             r82.0<1>:f  0x0:f // $54:&56:
-    // mov (16|M16)         r90.0<1>:f  0x0:f // $55:&57:
-    //
-    // Reuse r81, r89, r82, r90 in the gather send
-    auto getRemoveableImm = [this](G4_INST *inst,
-                                   std::vector<G4_INST *> &immMovs) {
-      // The instruction is only used for payload preparation.
-      if (inst->use_size() != 1) {
-        return (G4_INST *)nullptr;
-      }
-
-      if (inst->opcode() != G4_mov) {
-        return (G4_INST *)nullptr;
-      }
-
-      G4_DstRegRegion *dst = inst->getDst();
-      // dst GRF aligned and contigous
-      if (dst->getSubRegOff() || dst->getHorzStride() != 1) {
-        return (G4_INST *)nullptr;
-      }
-
-      if (kernel.fg.globalOpndHT.isOpndGlobal(dst)) {
-        return (G4_INST *)nullptr;
-      }
-
-      // GRF Alignment with physical register assigned
-      if (dst->getLinearizedStart() % builder.getGRFSize() != 0) {
-        return (G4_INST *)nullptr;
-      }
-
-      // If the destination operand size is less than 1 GRF
-      if ((dst->getLinearizedEnd() - dst->getLinearizedStart() + 1) <
-          builder.getGRFSize()) {
-        return (G4_INST *)nullptr;
-      }
-
-      G4_Operand *src = inst->getSrc(0);
-      int64_t imm = src->asImm()->getImm();
-      for (size_t i = 0; i < immMovs.size(); i++) {
-        G4_INST *imov = immMovs[i];
-        G4_Operand *isrc = imov->getSrc(0);
-        int64_t iimm = isrc->asImm()->getImm();
-        if (imm == iimm &&
-            src->getType() == isrc->getType() && // Same value and same type
-            inst->getDst()->getType() ==
-                imov->getDst()->getType() && // Same dst type
-            inst->getDst()->asDstRegRegion()->getHorzStride() ==
-                imov->getDst()
-                    ->asDstRegRegion()
-                    ->getHorzStride() &&                  // Same region
-            inst->getExecSize() == imov->getExecSize() && // Same execution size
-            inst->getMaskOffset() ==
-                imov->getMaskOffset()) { // Same mask offset
-          return imov;
-        }
-      }
-      immMovs.push_back(inst);
-
-      return (G4_INST *)nullptr;
-    };
 
     // The source opndNum of send instruction which was defined
     Gen4_Operand_Number opndNum = (*I).second;
@@ -691,6 +357,9 @@ bool SRSubPassAfterRA::isSRCandidateAfterRA(G4_INST *inst,
           });
       // if multiple defined, cannot be removed
       if (iter != dstSrcRegs.dstSrcMap.end()) {
+        if (!canRegisterBeEncoded(dst)) {
+          return false;
+        }
         for (unsigned offset = startOffset; offset < (startOffset + dstSize);
              offset++) {
           notRemoveableMap.push_back(std::make_pair(opndNum, offset));
@@ -703,6 +372,9 @@ bool SRSubPassAfterRA::isSRCandidateAfterRA(G4_INST *inst,
         movInstNum++;
       }
     } else {
+      if (!canRegisterBeEncoded(dst)) {
+        return false;
+      }
       if (movInst->opcode() == G4_mov && movInst->getSrc(0) &&
           movInst->getSrc(0)->isImm()) {
         // Check if there is mov instruction with same imm value
@@ -754,6 +426,58 @@ bool SRSubPassAfterRA::isSRCandidateAfterRA(G4_INST *inst,
   if (movInstNum < 2) {
     return false;
   }
+  // Even keep original GRF of src0 or src1, the original one cannot be more than 255
+  unsigned short GRFSize = builder.getGRFSize();
+  int src0Size = inst->getMsgDesc()->getSrc0LenRegs();
+  G4_Operand *src0 = inst->getSrc(0);
+  int j = 0;
+  for (int i = 0; i < src0Size; i++) {
+    bool replaced = false;
+    int srcOffset = src0->getLeftBound() / GRFSize + i;
+    if (j < (int)dstSrcRegs.dstSrcMap.size() &&
+        dstSrcRegs.dstSrcMap[j].opndNum == Opnd_src0) {
+      int opndSize =
+          (dstSrcRegs.dstSrcMap[j].opnd->getLinearizedEnd() -
+           dstSrcRegs.dstSrcMap[j].opnd->getLinearizedStart() + GRFSize - 1) /
+          GRFSize;
+      int opndOffset = dstSrcRegs.dstSrcMap[j].offset;
+
+      if ((srcOffset >= opndOffset) && (srcOffset < opndOffset + opndSize)) {
+        replaced = true;
+        i += opndSize - 1;
+        j++;
+      }
+    }
+    if (!replaced && (src0->getLinearizedStart() / GRFSize + i) > 255) {
+      return false;
+    }
+  }
+
+  G4_Operand *src1 = inst->getSrc(0);
+  if (src1 && !src1->isNullReg()) {
+    int src1Size = inst->getMsgDesc()->getSrc1LenRegs();
+    for (int i = 0; i < src1Size; i++) {
+      bool replaced = false;
+      int src1ffset = src1->getLeftBound() / GRFSize + i;
+      if (j < (int)dstSrcRegs.dstSrcMap.size() &&
+          dstSrcRegs.dstSrcMap[j].opndNum == Opnd_src1) {
+        int opndSize =
+            (dstSrcRegs.dstSrcMap[j].opnd->getLinearizedEnd() -
+             dstSrcRegs.dstSrcMap[j].opnd->getLinearizedStart() + GRFSize - 1) /
+            GRFSize;
+        int opndOffset = dstSrcRegs.dstSrcMap[j].offset;
+
+        if ((src1ffset >= opndOffset) && (src1ffset < opndOffset + opndSize)) {
+          replaced = true;
+          i += opndSize - 1;
+          j++;
+        }
+      }
+      if (!replaced && (src1->getLinearizedStart() / GRFSize + i) > 255) {
+        return false;
+      }
+    }
+  }
 
   dstSrcRegs.firstDefID = firstDefID;
   // Sort according to the register order in the original payload
@@ -763,6 +487,88 @@ bool SRSubPassAfterRA::isSRCandidateAfterRA(G4_INST *inst,
   return true;
 }
 
+bool SRSubPassAfterRA::checkCandidateForLargeGRF(G4_INST *inst,
+                                                 regCandidatesBRA &dstSrcRegs) {
+  // Check the source only, because large GRF is dst doesn't affect index
+  // register encoding.
+  // At the same time, much check each register of indexed regsiters. Because
+  // the original register may be replace, may not be replaced. So cannot use
+  // operand based check
+
+  unsigned short GRFSize = builder.getGRFSize();
+  int j = 0;
+  // Check the source of define mov if the mov can be removed, otherwise, check
+  // original source 0 directly
+  int src0Size = inst->getMsgDesc()->getSrc0LenRegs();
+  G4_Operand *src0 = inst->getSrc(0);
+  for (int i = 0; i < src0Size; i++) {
+    if (j < (int)dstSrcRegs.dstSrcMap.size() &&
+        dstSrcRegs.dstSrcMap[j].opndNum == Opnd_src0) {
+      int opndSize =
+          (dstSrcRegs.dstSrcMap[j].opnd->getLinearizedEnd() -
+           dstSrcRegs.dstSrcMap[j].opnd->getLinearizedStart() + GRFSize - 1) /
+          GRFSize;
+      int srcOffset = src0->getLeftBound() / GRFSize + i;
+      int opndOffset = dstSrcRegs.dstSrcMap[j].offset;
+
+      if ((srcOffset >= opndOffset) && (srcOffset < opndOffset + opndSize)) {
+        for (int k = 0; k < opndSize; k++) {
+          if ((dstSrcRegs.dstSrcMap[j].opnd->getLinearizedStart() +
+               k * GRFSize) > (256 * builder.numEltPerGRF<Type_UB>() - 1)) {
+            dstSrcRegs.isLargeGRF = true;
+          }
+        }
+        i += opndSize - 1;
+        j++;
+        continue;
+      }
+    }
+
+    // Not replaced
+    if ((src0->getLinearizedStart() + i * GRFSize) >
+        (256 * builder.numEltPerGRF<Type_UB>() - 1)) {
+      dstSrcRegs.isLargeGRF = true;
+    }
+  }
+
+  // Check the ource of define mov if the mov can be removed, otherwise, check
+  // original source 1 register
+  G4_Operand *src1 = inst->getSrc(1);
+  int src1Size = inst->getMsgDesc()->getSrc1LenRegs();
+  if (src1 && !src1->isNullReg()) {
+    for (int i = 0; i < src1Size; i++) {
+      if (j < (int)dstSrcRegs.dstSrcMap.size() &&
+          dstSrcRegs.dstSrcMap[j].opndNum == Opnd_src1) {
+        int opndSize =
+            (dstSrcRegs.dstSrcMap[j].opnd->getLinearizedEnd() -
+             dstSrcRegs.dstSrcMap[j].opnd->getLinearizedStart() + GRFSize - 1) /
+            GRFSize;
+        int srcOffset = src1->getLeftBound() / GRFSize + i;
+        int opndOffset = dstSrcRegs.dstSrcMap[j].offset;
+
+        if ((srcOffset >= opndOffset) && (srcOffset < opndOffset + opndSize)) {
+          for (int k = 0; k < opndSize; k++) {
+            if ((dstSrcRegs.dstSrcMap[j].opnd->getLinearizedStart() +
+                 k * GRFSize) > (256 * builder.numEltPerGRF<Type_UB>() - 1)) {
+              dstSrcRegs.isLargeGRF = true;
+            }
+          }
+          i += opndSize - 1;
+          j++;
+          continue;
+        }
+      }
+
+      // Add source 1 register directly
+      if ((src1->getLinearizedStart() + i * GRFSize) >
+          (256 * builder.numEltPerGRF<Type_UB>() - 1)) {
+        dstSrcRegs.isLargeGRF = true;
+      }
+    }
+  }
+
+  return !dstSrcRegs.isLargeGRF;
+}
 
 // Replace the send instruction with the payload of
 // Insert the scalar register intialization mov instructions.
@@ -983,6 +789,7 @@ void SRSubPassAfterRA::SRSubAfterRA(G4_BB *bb) {
       INST_LIST_RITER scan_ri = ri;
       scan_ri++;
       G4_INST *rInst = *scan_ri;
+      bool unsupportGRF = false;
 
       while (rInst->getLocalId() > candidates[inst].firstDefID) {
         if (rInst->isDead()) {
@@ -1040,6 +847,13 @@ void SRSubPassAfterRA::SRSubAfterRA(G4_BB *bb) {
               int srcRegLB = (*dstSrcRegsIter).opnd->getLinearizedStart();
               int srcRegRB = (*dstSrcRegsIter).opnd->getLinearizedEnd();
               if (!(srcRegRB < dstRegLB || srcRegLB > dstRegRB)) {
+                // mov is not removeable and dst register > r255, cannot be
+                // gather instruction.
+                G4_DstRegRegion *removedDst = (*dstSrcRegsIter).inst->getDst();
+                if (!canRegisterBeEncoded(removedDst)) {
+                  unsupportGRF = true;
+                  break;
+                }
 
                 // Register is redefined
                 dstSrcRegsIter =
@@ -1050,6 +864,9 @@ void SRSubPassAfterRA::SRSubAfterRA(G4_BB *bb) {
             }
           }
         }
+        if (unsupportGRF) {
+          break;
+        }
         scan_ri++;
         if (scan_ri == rend) {
           break;
@@ -1057,6 +874,10 @@ void SRSubPassAfterRA::SRSubAfterRA(G4_BB *bb) {
         rInst = *scan_ri;
       }
 
+      if (unsupportGRF || !checkCandidateForLargeGRF(inst, candidates[inst])) {
+        candidates.erase(candidatesIt);
+        continue;
+      }
       // Due to extra mov for s0, so don't use s0 if equal or less than 1 mov
       // inst can be removed.
       if (candidates[inst].dstSrcMap.size() <= 1 &&
